import type { NamespaceTable } from "./NameSpaceTable";
import { coerceNodeId, NodeId } from "./NodeId";
import { UAReference } from "./UAReference";
import { assert } from "@/util/assert";
import { XMLElem, type IToXML } from "@/util/XmlElem";

export class UABaseNode implements IToXML{
    constructor(public nodeId: NodeId,
                public browseName: string,
                public displayName: string,
                public references: UAReference[]) {
    }

    public static nullBaseNode=new UABaseNode(NodeId.nullNodeId,"","",[]);

    reIndex(nst: NamespaceTable, onst: NamespaceTable) {
        const nsName=onst.getUri(this.nodeId.namespace);
        assert(nsName!=undefined)
        const newIndex=nst.getIndex(nsName);
        assert(newIndex!=undefined)
        this.nodeId.namespace=newIndex;

        for(const uaref of this.references) {
            uaref.reIndex(nst, onst);
        }
    }

    getChildren(): UABaseNode[] {
        const children: UABaseNode[]=[];
        for(const ref of this.references) {
            switch(ref.referenceType) {
                case 'HasComponent':
                case 'HasOrderedComponent':
                case 'Organizes':
                case 'HasProperty':
                case 'HasSubtype':
                case 'HasAddIn':
                    if(ref.isForward&&ref.fromNode===this) {
                        if(!children.includes(ref.toNode))
                            children.push(ref.toNode);
                    } 
                    if(!ref.isForward&&ref.toNode===this) {
                        if(!children.includes(ref.fromNode))
                            children.push(ref.fromNode);
                    }
                    break;
                }
        }
        return children;
    }

    resolveReferences(nm: Map<string, UABaseNode>) {
        for(const ref of this.references) {
            const fromNode=nm.get(ref.fromRef.toString()) 
            if(fromNode)
                ref.fromNode=fromNode;
            const toNode=nm.get(ref.toRef.toString()) 
            if(toNode) //TODO: if we cant find the node; the parser is incomplete or the nodeset is broken
                ref.toNode=toNode;
            if(ref.fromNode===this&&toNode){ //add this reference to referenced node
                if(!ref.toNode.references.includes(ref))
                    ref.toNode.references.push(ref);
            } 
        }
    }

    static fromXML(xmlObject: any): UABaseNode{
        const xmlReferences=xmlObject['References'];
        const references:UAReference[]=[];
        const nodeId=coerceNodeId(xmlObject['@_NodeId']);
        for(const xmlref of xmlReferences.Reference) {
            references.push(UAReference.fromXML(xmlref, nodeId));
        }
        const ua=new UABaseNode(nodeId, xmlObject['@_BrowseName'], xmlObject['DisplayName']['#text'], references);
        return ua;
    }

    toXML(): XMLElem {
        throw new Error("UABaseNode has no xml rep; implement in subtype.");
    }
}